"""
Tasks.

Background tasks that run a worker server.
These tasks can run even while the user is not online,
not making any active HTTP requests, or not using the application.
"""

import typing
from enum import IntEnum

import celery
import celery.schedules

from sap.loggers import logger
from sap.settings import SapSettings


class FetchStrategy(IntEnum):
    """Define if new or old data should be fetched."""

    NEW: int = 1
    OLD: int = 2


class CronResponse(typing.TypedDict, total=False):
    """Define a standard cron task response."""

    error: dict[str, str]
    result: dict[str, int]


class CronTask(celery.Task):
    """Define how cron task classes should be structured."""

    expires = 60 * 60  # automatically expires if not run within 1 hour
    time_limit = 60 * 60 * 3  # default to 3 hours, automatically kill the task if exceed the limit
    args: list[typing.Any]
    kwargs: dict[str, typing.Any]
    schedule: celery.schedules.crontab

    def __init__(self, **kwargs: typing.Any) -> None:
        """Initialize the cron task."""
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def get_name(cls) -> str:
        """Get Name of the current Task."""
        return cls.__module__.split(".crons", maxsplit=1)[0] + "." + str(cls.__name__)

    def get_queryset(self, *, batch_size: typing.Optional[int] = None, **kwargs: typing.Any) -> typing.Any:
        """Fetch the list of elements to process."""
        raise NotImplementedError

    async def process(self, *, batch_size: int = 100, **kwargs: typing.Any) -> typing.Any:
        """Run the cron task and process elements."""
        raise NotImplementedError

    async def get_stats(self) -> dict[str, int]:
        """Give stats about the number of elements left to process."""
        raise NotImplementedError

    async def run(self) -> CronResponse:
        """Run the task and save meta info to Airtable."""
        response: CronResponse

        # B. Runs the task
        try:
            result = await self.process(**self.kwargs)
        except Exception as exc:  # pylint: disable=broad-except;
            if not SapSettings.is_env_prod:
                raise
            logger.exception(exc)
            response = {"error": {"class": exc.__class__.__name__, "message": str(exc)}}
        else:
            response = {"result": result}

        return response


def register_crontask(
    crontask_class: type[CronTask],
    schedule: celery.schedules.crontab,
    kwargs: typing.Optional[dict[str, typing.Any]] = None,
) -> CronTask:
    """Register a task on the worker servers."""
    return crontask_class(schedule=schedule, kwargs=kwargs or {})



"""
Models.

Base models that can be sub-classed independently by each Trellis integration.
"""
from __future__ import annotations

import typing
from datetime import datetime

import beanie
import pydantic

if typing.TYPE_CHECKING:
    T = typing.TypeVar("T")

    class Link(beanie.Link[T]):
        """Fix typing issue mypy when querying related fields."""

        id: beanie.PydanticObjectId

else:
    from beanie import Link


class _DocMeta(pydantic.BaseModel):
    """Meta Data allowing to keep trace of Documents versioning and updates."""

    version: int = 0  # version of the document being imported
    source: typing.Optional[str]  # where the data is coming from: webhook, cron
    created: typing.Optional[datetime]  # when the document was first imported
    updated: typing.Optional[datetime]  # when the document was last updated
    deleted: typing.Optional[datetime]  # when the document was deleted, (deleted document may be retained for logging)


class DocMeta(pydantic.BaseModel):
    """Manage meta data and ensure that it's correctly set."""

    doc_meta: _DocMeta = _DocMeta()

    @pydantic.root_validator
    @classmethod
    def validate_doc_meta(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """Validate doc meta on each model update."""
        doc_meta: _DocMeta = values["doc_meta"]
        doc_meta.updated = datetime.now()
        doc_meta.created = doc_meta.created or doc_meta.updated
        return values



import asyncio
import json
import time

import aioamqp.channel
import aioamqp.envelope
import aioamqp.exceptions
import aioamqp.properties
import pytest

from AppMain.settings import AppSettings
from sap.settings import SapSettings
from sap.worker import AMQPClient, SignalPacket

AMQPClient.db_params = AppSettings.RABBITMQ


@pytest.mark.asyncio
async def test_signal_packet_queue_declaration() -> None:
    """Test that queues are declared successfully"""
    packet = SignalPacket("sap.#", providing_args=["identifier"])
    task_name = "tests.LambdaWorker"

    channel = await packet.get_default_channel()

    queue_primary_name = packet.queue_get_name(task_name=task_name)
    queue_fallback_name = packet.queue_get_name(task_name=task_name, is_fallback=True)

    async def check_queue_exits(channel_: aioamqp.channel.Channel, queue_name: str, result: bool) -> None:
        try:
            await channel_.queue_declare(queue_name=queue_name, passive=True)
        except aioamqp.exceptions.ChannelClosed:
            assert not result, "Checking channel existence return `False`, expecting `True`"
            await channel_.open()
        else:
            assert result, "Checking channel existence return `True`, expecting `False`"

    # A. Ensure that all queues have been deleted
    await channel.queue_delete(queue_name=queue_primary_name)
    await check_queue_exits(channel, queue_primary_name, result=False)
    await channel.queue_delete(queue_name=queue_fallback_name)
    await check_queue_exits(channel, queue_fallback_name, result=False)

    # reset connection
    # await channel.close()
    packet.connections.pop("default")

    # B. Run queue declaration
    await packet.queue_declare(task_name=task_name)

    # Ensure that the queues have been created and binding successfully
    await packet.connection_reset()
    channel = await packet.get_default_channel()
    await check_queue_exits(channel, queue_primary_name, result=True)
    await check_queue_exits(channel, queue_fallback_name, result=True)

    # Other: Test that reset connection works when channel is already closed
    await channel.close()
    await packet.connection_reset()


@pytest.mark.asyncio
async def test_signal_packet_send() -> None:
    """Testing that sending messages is working."""
    SapSettings.is_env_dev = False
    task_name = "tests.LambdaWorker"
    identifier = "card_12345"
    timestamp = int(time.time())
    result = {}

    # Create received queue
    packet_receiver = SignalPacket("sap.#", providing_args=["identifier"])
    await packet_receiver.queue_declare(task_name=task_name)

    # Create queue consumer client
    async def callback_receiver(
        channel: aioamqp.channel.Channel,
        payload: bytes,
        envelop: aioamqp.envelope.Envelope,
        properties: aioamqp.properties.Properties,
    ) -> None:
        # print(f'Message received {payload=} {properties=}')
        await channel.basic_client_ack(envelop.delivery_tag)
        result["receiver"] = json.loads(payload.decode("utf-8"))
        result["properties"] = properties

    channel = await packet_receiver.get_default_channel()
    await channel.basic_consume(callback_receiver, packet_receiver.queue_get_name(task_name=task_name), no_wait=True)

    # Send messages
    packet_sender = SignalPacket(f"sap.app.{identifier}.user.created", providing_args=["identifier"])
    await packet_sender.send(identifier=identifier, timestamp=timestamp)

    await asyncio.sleep(1)

    # Check that messages has been received
    assert "receiver" in result
    assert result["receiver"]["identifier"] == identifier
    assert result["receiver"]["kwargs"]["timestamp"] == timestamp

    # Test that packet are disable by default in dev
    SapSettings.is_env_dev = True
    await packet_sender.send(identifier=identifier, timestamp=timestamp)



import asyncio
import time
import typing

import celery
import celery.worker
import pytest

from AppMain.settings import AppSettings
from sap.worker import AMQPClient, LambdaTask, LambdaWorker, SignalPacket, register_lambda

AMQPClient.db_params = AppSettings.RABBITMQ


class DummyLambdaTask(LambdaTask):
    """Create dummy lambda task class to ensure that LambdaTask class is functioning."""

    results: list[int] = []

    packet = SignalPacket("sap_tests.app.*.user.created", providing_args=["identifier", "timestamp"])

    async def handle_receive(self, *args: str, **kwargs: typing.Any) -> dict[str, typing.Any]:
        self.results.append(kwargs["timestamp"])
        return {"result": True}


def test_lambda_task() -> None:
    """Create dummy lambda task to ensure that LambdaTask class is functioning."""
    task = register_lambda(DummyLambdaTask)
    result = task.run(timestamp=1)
    assert "result" in result


class DummyLambdaWorker(LambdaWorker):
    """Create a dummy Lambda worker for testing."""

    packets = [SignalPacket("sap_tests.#", providing_args=["identifier", "kwargs"])]
    name = "tests.LambdaWorker"

    def get_task_list(self) -> list[LambdaTask]:
        """Register dummy task."""
        return [register_lambda(DummyLambdaTask)]


@pytest.fixture
def setup_celery_app(celery_app: celery.Celery) -> None:
    """Setting up Celery worker"""
    celery_app.register_task(register_lambda(DummyLambdaTask))
    celery_app.steps["consumer"].add(DummyLambdaWorker)


@pytest.mark.asyncio
async def test_lambda_worker(setup_celery_app: None, celery_worker: celery.worker.WorkController) -> None:
    """Create dummy lambda worker to ensure that LambdaWorker class is functioning."""
    identifier = "card_12345"
    timestamp = int(time.time())

    # Send packet
    packet_yes = SignalPacket(f"sap_tests.app.{identifier}.user.created", providing_args=["identifier", "timestamp"])
    packet_no = SignalPacket(f"sap_tests.app.{identifier}.merchant.updated", providing_args=["identifier", "timestamp"])
    # print(f"---> Sending packet {identifier=} {timestamp=}")
    await packet_yes.send(identifier, timestamp=timestamp + 1)
    await packet_no.send(identifier, timestamp=timestamp + 2)
    await packet_yes.send(identifier, timestamp=timestamp + 3)
    await packet_no.send(identifier, timestamp=timestamp + 4)

    await asyncio.sleep(3)

    assert timestamp + 1 in DummyLambdaTask.results
    assert timestamp + 2 not in DummyLambdaTask.results
    assert timestamp + 3 in DummyLambdaTask.results
    assert timestamp + 4 not in DummyLambdaTask.results



"""
Auth.

Utils to handle user authentication and permissions.
"""
import base64
import binascii
import time
import typing

import jwt

from fastapi import Cookie, Depends, Request, Response, status
from fastapi.exceptions import HTTPException
from starlette.authentication import AuthCredentials, AuthenticationBackend, AuthenticationError

from AppMain.settings import AppSettings
from sap.beanie import Document


class JWTAuth:
    """JWT cookie authentication utils.

    This can be used to login/logout user with persistent sessions.
    Mainly useful for Web Apps. For API, it is better to use Basic Authentication in Headers.
    """

    auth_login_url: typing.ClassVar[str] = "/pages/auth/login/"
    auth_cookie_key: typing.ClassVar[str] = "user_session"
    auth_cookie_expires: typing.ClassVar[int] = 60 * 60 * 12  # expiration = 12 hours

    @classmethod
    def get_auth_login_url(cls) -> str:
        """Retrieve the login url where user are redirect in case of auth failure."""
        return cls.auth_login_url

    @classmethod
    def get_auth_cookie_key(cls) -> str:
        """Retrieve key used to define the authentication cookie."""
        return cls.auth_cookie_key

    @classmethod
    def get_auth_cookie_expires(cls) -> int:
        """Retrieve validity in seconds of the authentication cookie."""
        return cls.auth_cookie_expires

    @classmethod
    def create_token(cls, user: Document, expires: typing.Optional[int] = None) -> str:
        """Get JWT temporary token."""
        expires = cls.get_auth_cookie_expires() if expires is None else expires
        jwt_data = {"exp": int(time.time()) + expires, "user_id": str(user.id)}
        return jwt.encode(payload=jwt_data, key=AppSettings.CRYPTO_SECRET, algorithm="HS256")

    @classmethod
    async def find_user(cls, jwt_token: str, user_model_class: type[Document]) -> Document:
        """
        Verify that JWT token is valid.

        :param jwt_token: The lifespan of the token in seconds.
        :param user_model_class: user model class
        """
        jwt_data = jwt.decode(jwt_token, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
        # Raises: jwt.exceptions.InvalidTokenError => Token has expired or is invalid

        user = await user_model_class.find_one_or_404(
            user_model_class.id == jwt_data["user_id"],
            user_model_class.is_active == True,
        )
        # Raise: Object404Error => User cannot be found

        return user

    @classmethod
    async def login(cls, response: Response, user: Document) -> Response:
        """Create a persistent cookie based session for the authenticated user."""
        response.set_cookie(key=cls.get_auth_cookie_key(), value=cls.create_token(user=user), httponly=True)
        return response

    @classmethod
    async def logout(cls, response: Response) -> Response:
        """Create a persistent cookie based session for the authenticated user."""
        response.delete_cookie(key=cls.get_auth_cookie_key(), httponly=True)
        return response

    @classmethod
    def depends(cls, user_model_class: type[Document]) -> typing.Any:
        """Provide the authenticated user to views that require it."""

        async def retrieve_user(
            request: Request, jwt_cookie: str = Cookie(default="", alias=cls.get_auth_cookie_key())
        ) -> "Document":
            user = None
            if jwt_cookie:
                try:
                    jwt_data = jwt.decode(jwt_cookie, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
                except jwt.exceptions.InvalidTokenError:
                    pass
                else:
                    user = await user_model_class.get(jwt_data["user_id"])
            if not user:
                raise HTTPException(
                    status_code=status.HTTP_307_TEMPORARY_REDIRECT, headers={"Location": cls.get_auth_login_url()}
                )
            return user

        return Depends(retrieve_user)


class JWTAuthBackend(AuthenticationBackend, JWTAuth):
    """Starlette Backend to authenticate use through JWT Token in Cookies."""

    user_model: type[Document]

    def __init__(self, user_model: type[Document]) -> None:
        """Initialize the authentication backend.

        :param user_model_class: The User model.
        """
        super().__init__()
        self.user_model = user_model

    async def authenticate(self, conn)-> typing.Optional[typing.Tuple["AuthCredentials", "BaseUser"]]:
        """Authenticate the user using Cookies."""
        if self.get_auth_cookie_key() not in conn.cookies:
            return

        jwt_cookie = conn.cookies[self.get_auth_cookie_key()]

        try:
            jwt_data = jwt.decode(jwt_cookie, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
        except jwt.exceptions.InvalidTokenError:
            raise AuthenticationError("Invalid JWT token")
        else:
            user = await self.user_model.get(jwt_data["user_id"])

        return AuthCredentials([user.get_scopes()]), user


class BasicAuthBackend(AuthenticationBackend):
    """Starlette Backend to authenticate use through Basic Token in Header."""

    user_model: type[Document]
    auth_key_attribute: str

    def __init__(self, user_model: type[Document], auth_key_attribute: str = "auth_key") -> None:
        """Initialize the authentication backend.

        :param user_model_class: The User model
        :param auth_key_attribute: The authentication key attribute on the User model
        """
        super().__init__()
        self.user_model = user_model
        self.auth_key_attribute = auth_key_attribute

    async def authenticate(self, conn):
        """Authenticate the user use the Authorization headers."""
        if "Authorization" not in conn.headers:
            return

        auth = conn.headers["Authorization"]
        try:
            scheme, credentials = auth.split()
            if scheme.lower() != "basic":
                return
            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error):
            raise AuthenticationError("Invalid basic auth credentials")

        username, _, pwd = decoded.partition(":")

        auth_key = getattr(self.user_model, self.auth_key_attribute)
        user = await self.user_model.find_one_or_404(auth_key == (username or pwd))

        return AuthCredentials(user.get_scopes()), user



"""
Beanie Client.

Initialize connection to the Mongo Database.
"""
import typing

from motor.motor_asyncio import AsyncIOMotorClient

import beanie

from sap.settings import DatabaseParams


class BeanieClient:
    """Set up a connection to the MongoDB server."""

    connections: typing.ClassVar[dict[str, AsyncIOMotorClient]] = {}

    @classmethod
    async def get_db_default(cls) -> AsyncIOMotorClient:
        """Return the default db connection."""
        return cls.connections["default"]

    @classmethod
    async def init(
        cls,
        mongo_params: DatabaseParams,
        document_models: list[typing.Union[typing.Type[beanie.Document], typing.Type[beanie.View], str]],
    ) -> None:
        """Open and maintain a connection to the database.

        :force bool: Use it for force a connection initialization
        """
        # TODO: Check if connection exist, return existing connection
        if "default" in cls.connections:
            # connection = cls.connections["default"]
            return

        client = AsyncIOMotorClient(mongo_params.get_dns())
        connection = client[mongo_params.db]
        cls.connections["default"] = connection
        await beanie.init_beanie(connection, document_models=document_models)
        print("--> Initialized beanie")



"""
# Serializers.

Handle data validation.
"""
from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar, Union

from fastapi import Request
from pydantic import BaseModel
from pydantic.fields import SHAPE_LIST, ModelField

from sap.beanie.document import Document, TDoc

from . import utils

# SerializerType = TypeVar("SerializerType", bound=BaseModel)

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny


class CursorInfo:
    """Contains information on how the list should paginated."""

    offset: int = 0
    limit: int = 10
    sort: str = "-doc_meta.created"

    def __init__(self, request: Request) -> None:
        """Initialize the cursor info."""
        cursor_str = request.query_params.get("cursor", "")
        try:
            limit, offset = utils.base64_url_decode(cursor_str).split(",")
        except ValueError:
            return
        self.limit, self.offset = int(limit), int(offset)

    def get_beanie_query_params(self) -> dict[str, Any[int, str]]:
        """Return params to apply to the database query when using beanie."""
        return {
            "limit": self.limit,
            "skip": self.offset,
            "sort": self.sort,
        }

    def get_next(self) -> Optional[str]:
        """Get the cursor to paginate forward."""
        offset = self.offset + self.limit
        return utils.base64_url_encode(f"{self.limit},{offset}")

    def get_previous(self) -> Optional[str]:
        """Get the cursor to paginate backward."""
        offset = self.offset - self.limit
        if offset <= 0:
            return None
        return utils.base64_url_encode(f"{self.limit},{offset}")


class ObjectSerializer(Generic[TDoc], BaseModel):
    """Serialize an object for retrieve or list."""

    @classmethod
    def get_id(cls, instance: TDoc) -> str:
        """Return the Mongo ID of the object."""
        return str(instance.id)

    @classmethod
    def get_created(cls, instance: TDoc) -> datetime.datetime:
        """Return the user creation date."""
        assert instance.doc_meta.created  # let mypy know that this cannot be null
        return instance.doc_meta.created

    @classmethod
    def get_updated(cls, instance: TDoc) -> datetime.datetime:
        """Return the user creation date."""
        assert instance.doc_meta.updated  # let mypy know that this cannot be null
        return instance.doc_meta.updated

    @classmethod
    def _get_instance_data(
        cls, instance: TDoc, exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None
    ) -> dict[str, Any]:
        """Retrieve the serializer value from the instance and getters."""
        data = {}
        exclude = exclude or set()
        for field_name, field in cls.__fields__.items():
            if field_name in exclude:
                continue
            elif hasattr(cls, f"get_{field_name}"):
                data[field_name] = getattr(cls, f"get_{field_name}")(instance=instance)
            elif issubclass(field.type_, ObjectSerializer):
                related_object = getattr(instance, field_name, None)
                if field.shape == SHAPE_LIST:
                    data[field_name] = (
                        field.type_.read_list(related_object, exclude=field.field_info.exclude)
                        if related_object
                        else []
                    )
                else:
                    data[field_name] = (
                        field.type_.read(related_object, exclude=field.field_info.exclude) if related_object else None
                    )
            else:
                data[field_name] = getattr(instance, field_name)
        return data

    @classmethod
    def read(cls, instance: TDoc, exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None) -> "SerializerType":
        """Serialize a single object instance."""
        return cls(**cls._get_instance_data(instance, exclude=exclude))

    @classmethod
    def read_list(
        cls, instance_list: list[TDoc], exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None
    ) -> list["SerializerType"]:
        """Serialize a list of objects."""
        return [cls.read(instance, exclude=exclude) for instance in instance_list]

    @classmethod
    def read_page(
        cls,
        instance_list: list[TDoc],
        cursor_info: CursorInfo,
        request: Request,
    ) -> PaginatedData["SerializerType"]:
        """Serialize a list of objects."""

        # TODO: implemented proper cursor pagination, for now fake it till you make it.

        next = cursor_info.get_next()
        previous = cursor_info.get_previous()
        return PaginatedData(
            count=0,
            next=str(request.url.include_query_params(cursor=next)) if next else None,
            previous=str(request.url.include_query_params(cursor=previous)) if previous else None,
            data=cls.read_list(instance_list),
        )


SerializerType = TypeVar("SerializerType", bound=ObjectSerializer)


class PaginatedData(Generic[SerializerType], BaseModel):
    """Represent the structure of an API paginated list response."""

    object: str = "list"
    count: int
    next: Optional[str]
    previous: Optional[str]
    data: list[Any]


class WriteObjectSerializer(Generic[TDoc], BaseModel):
    """Serialize an object for create or update."""

    instance: Optional[TDoc] = None

    async def run_async_validators(self) -> None:
        """Check that data pass DB validation."""

        field: ModelField
        embedded_serializers = {}
        for field_name, field in self.__fields__.items():
            if issubclass(field.type_, WriteObjectSerializer):
                embedded_serializers[field_name] = field

        field_serializer: WriteObjectSerializer[TDoc]
        for field_name in embedded_serializers:
            if field_serializer := getattr(self, field_name):
                if self.instance:
                    field_serializer.instance = getattr(self.instance, field_name)
                await field_serializer.run_async_validators()

    def dict(
        self,
        *,
        include: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        exclude: Optional[Union["AbstractSetIntStr", "MappingIntStrAny"]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> "DictStrAny":
        """Dump the serializer data with exclusion of unwanted fields."""
        # Exclude from dumping
        exclude = (exclude or {}) | {"instance": True}

        # Some fields are only excluded from being cascade dumps to dict,
        # but their original value is still needed
        exclude_doc_dumps = {}

        # Embedded documents need to be converted to object after dumps
        embedded_serializers = {}

        for field_name, field in self.__fields__.items():
            if field_name in exclude:
                continue

            if issubclass(field.type_, Document):
                exclude_doc_dumps[field_name] = True

            if issubclass(field.type_, WriteObjectSerializer):
                embedded_serializers[field_name] = field.type_.__fields__["instance"].type_

            # Some fields are excluded as they are only needed for create
            if field.field_info.extra.get("exclude_update") and self.instance:
                exclude[field_name] = True

            # Some fields are excluded as they are only needed for update
            if field.field_info.extra.get("exclude_create") and not self.instance:
                exclude[field_name] = True

        exclude = exclude | exclude_doc_dumps

        result = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        for field_name in exclude_doc_dumps:
            result[field_name] = getattr(self, field_name)

        instance_embedded: BaseModel
        for field_name, field_model in embedded_serializers.items():
            if not result[field_name]:
                continue
            if instance_embedded := getattr(self.instance, field_name, None):
                result[field_name] = instance_embedded.copy(update=result[field_name])
            else:
                result[field_name] = field_model(**result[field_name])

        return result

    async def create(self, **kwargs: Any) -> TDoc:
        """Create the object in the database using the data extracted by the serializer."""
        instance_class: type[TDoc] = self.__fields__["instance"].type_
        self.instance = await instance_class(**self.dict()).create()
        return self.instance

    async def update(self, **kwargs: Any) -> TDoc:
        """Update the object in the database using the data extracted by the serializer."""
        assert self.instance
        self.instance = self.instance.copy(update=self.dict())
        await self.instance.save()
        return self.instance
